// A generic key-to-value dictionary represented as a binary search tree.
#pragma once

#include "Pair.h"
#include "Vector.h"

#include <initializer_list>

namespace dictionary {

template <typename K, typename V>
class Dict {
public:
    using value_type = Pair<const K, V>;

    // Default constructor:
    Dict() = default;

    // Construct from a list of pairs:
    Dict(std::initializer_list<value_type>);

    // Construct from a pair of iterators to pairs.
    template <typename FwdIter>
    Dict(FwdIter start, FwdIter limit);

    // Copy constructor:
    Dict(const Dict&);

    // Copy-assignment operator:
    Dict& operator=(const Dict&);

    // Frees the tree's memory when we're done with it:
    ~Dict();

    bool empty() const;
    size_t size() const;

    // Removes all mappings from the tree
    void clear();

    // Does the dictionary contain the key?
    bool has_key(const K&) const;

    // Returns a pointer to the value associated with the given key, or
    // nullptr if not found.
    const V* find(const K&) const;
    V* find(const K&);

    // Inserts a mapping from the given key to the given value.
    void insert(const K&, const V&);

    // Lookup, but inserts V() (default V) if not found.
    V& operator[](const K&);

    Vector<K> get_keys() const;

    class iterator;
    class const_iterator;

    iterator begin();
    iterator end();

    const_iterator begin() const;
    const_iterator end() const;

    const_iterator cbegin() const;
    const_iterator cend() const;

private:
    struct Node;

    Node* root_  = nullptr;
    size_t size_ = 0;
};

template <typename K, typename V>
class Dict<K, V>::iterator
{
public:
    iterator() = default;

    value_type& operator*() const;
    value_type* operator->() const;
    iterator& operator++();
    iterator operator++(int);

    bool operator==(const iterator&) const;
    bool operator!=(const iterator&) const;

private:
    friend class Dict;
    friend class const_iterator;

    iterator(Node*);
    Vector<Node*> stack_;
};

template <typename K, typename V>
class Dict<K, V>::const_iterator
{
public:
    const_iterator() = default;
    const_iterator(const iterator&);

    const value_type& operator*() const;
    const value_type* operator->() const;
    const_iterator& operator++();
    const_iterator operator++(int);

    bool operator==(const const_iterator&) const;
    bool operator!=(const const_iterator&) const;

private:
    friend class Dict<K, V>;

    const_iterator(const Node*);
    Vector<const Node*> stack_;
};


///
/// IMPLEMENTATION
///

template <typename K, typename V>
struct Dict<K, V>::Node
{
    value_type data;
    Node* left;
    Node* right;

    Node(const value_type& kv, Node* l, Node* r)
            : data(kv), left(l), right(r)
    { }

    static void deep_delete(Node*);
    static Node* deep_clone(const Node*);

    static void get_keys_helper(Node*, Vector<K>&);

    static Node* find(Node*, const K&);
    static Node** find_mut(Node**, const K&);
};

template <typename K, typename V>
Dict<K, V>::Dict(std::initializer_list<value_type> mappings)
        : Dict(mappings.begin(), mappings.end())
{ }

template<typename K, typename V>
template<typename FwdIter>
Dict<K, V>::Dict(FwdIter start, FwdIter limit)
{
    for (auto i = start; i != limit; ++i) {
        insert(i->first, i->second);
    }
}

template <typename K, typename V>
Dict<K, V>::Dict(const Dict& that)
        : size_(that.size_),
          root_(Node::deep_clone(that.root_))
{ }

template <typename K, typename V>
Dict<K, V>& Dict<K, V>::operator=(const Dict& that)
{
    Node::deep_delete(root_);
    size_ = that.size_;
    root_ = Node::deep_clone(that.root_);

    return *this;
}

template <typename K, typename V>
Dict<K, V>::~Dict()
{
    Node::deep_delete(root_);
}

template <typename K, typename V>
bool Dict<K, V>::empty() const {
    return root_ == nullptr;
}

template <typename K, typename V>
size_t Dict<K, V>::size() const {
    return size_;
}

template <typename K, typename V>
void Dict<K, V>::clear()
{
    Node::deep_delete(root_);
    root_ = nullptr;
    size_ = 0;
}

template <typename K, typename V>
bool Dict<K, V>::has_key(const K& key) const
{
    return Node::find(root_, key) != nullptr;
}

template <typename K, typename V>
const V* Dict<K, V>::find(const K& key) const
{
    Node* node = Node::find(root_, key);
    if (node == nullptr) return nullptr;
    else return &node->data.second;
}

template <typename K, typename V>
V* Dict<K, V>::find(const K& key)
{
    Node* node = Node::find(root_, key);
    if (node == nullptr) return nullptr;
    else return &node->data.second;
}

template <typename K, typename V>
void Dict<K, V>::insert(const K& key, const V&value)
{
    Node** node = Node::find_mut(&root_, key);

    if (*node == nullptr) {
        *node = new Node({key, value}, nullptr, nullptr);
        ++size_;
    } else {
        (*node)->data.second = value;
    }
}

template <typename K, typename V>
V& Dict<K, V>::operator[](const K& key)
{
    Node** node = Node::find_mut(&root_, key);

    if (*node == nullptr) {
        *node = new Node({key, V()}, nullptr, nullptr);
        ++size_;
    }

    return (*node)->data.second;
}

template <typename K, typename V>
void Dict<K, V>::Node::get_keys_helper(Node* node, Vector<K>& out)
{
    if (node != nullptr) {
        Node::get_keys_helper(node->left, out);
        out.push_back(node->data.first);
        Node::get_keys_helper(node->right, out);
    }
}

template <typename K, typename V>
Vector<K> Dict<K, V>::get_keys() const
{
    Vector<K> result;
    Node::get_keys_helper(root_, result);
    return result;
}

template<typename K, typename V>
auto Dict<K, V>::begin() -> iterator
{
    return iterator(root_);
}

template<typename K, typename V>
auto Dict<K, V>::end() -> iterator
{
    return iterator();
}

template<typename K, typename V>
auto Dict<K, V>::begin() const -> const_iterator
{
    return const_iterator(root_);
}

template<typename K, typename V>
auto Dict<K, V>::end() const -> const_iterator
{
    return const_iterator();
}

template<typename K, typename V>
auto Dict<K, V>::cbegin() const -> const_iterator
{
    return const_iterator(root_);
}

template<typename K, typename V>
auto Dict<K, V>::cend() const -> const_iterator
{
    return const_iterator();
}

template <typename K, typename V>
void Dict<K, V>::Node::deep_delete(Node* node)
{
    if (node) {
        Node::deep_delete(node->left);
        Node::deep_delete(node->right);
        delete node;
    }
}

template <typename K, typename V>
auto Dict<K, V>::Node::deep_clone(const Node* node) -> Node*
{
    if (node)
        return new Node({node->data.first, node->data.second},
                        Node::deep_clone(node->left),
                        Node::deep_clone(node->right));
    else
        return nullptr;
}

template <typename K, typename V>
auto Dict<K, V>::Node::find(Node* node, const K& key) -> Node*
{
    while (node != nullptr) {
        if (key < node->data.first)
            node = node->left;
        else if (key > node->data.first)
            node = node->right;
        else
            break;
    }

    return node;
}

template <typename K, typename V>
auto Dict<K, V>::Node::find_mut(Node** node, const K& key) -> Node**
{
    while (*node != nullptr) {
        if (key < (*node)->data.first)
            node = &(*node)->left;
        else if (key > (*node)->data.first)
            node = &(*node)->right;
        else
            break;
    }

    return node;
}

template<typename K, typename V>
auto Dict<K, V>::iterator::operator*() const -> value_type&
{
    return stack_.back()->data;
}

template<typename K, typename V>
auto Dict<K, V>::iterator::operator->() const -> value_type*
{
    return &stack_.back()->data;
}

template<typename K, typename V>
auto Dict<K, V>::iterator::operator++() -> iterator&
{
    auto next_node = stack_.back();
    stack_.pop_back();
    if (next_node->left) stack_.push_back(next_node->left);
    if (next_node->right) stack_.push_back(next_node->right);
    return *this;
}

template<typename K, typename V>
auto Dict<K, V>::iterator::operator++(int) -> iterator
{
    auto old = *this;
    ++*this;
    return old;
}

template<typename K, typename V>
bool Dict<K, V>::iterator::operator==(const iterator& other) const
{
    return stack_ == other.stack_;
}

template<typename K, typename V>
bool Dict<K, V>::iterator::operator!=(const iterator& other) const
{
    return stack_ != other.stack_;
}

template<typename K, typename V>
Dict<K, V>::iterator::iterator(Node* start) : stack_{start}
{ }

template<typename K, typename V>
Dict<K, V>::const_iterator::const_iterator(const iterator& iter)
{
    stack_.reserve(iter.stack_.size());
    for (Node* node : iter.stack_)
        stack_.push_back(node);
}

template<typename K, typename V>
auto Dict<K, V>::const_iterator::operator*() const -> const value_type&
{
    return stack_.back()->data;
}

template<typename K, typename V>
auto Dict<K, V>::const_iterator::operator->() const -> const value_type*
{
    return &stack_.back()->data;
}

template<typename K, typename V>
auto Dict<K, V>::const_iterator::operator++() -> const_iterator&
{
    auto next_node = stack_.back();
    stack_.pop_back();
    if (next_node->left) stack_.push_back(next_node->left);
    if (next_node->right) stack_.push_back(next_node->right);
    return *this;
}

template<typename K, typename V>
auto Dict<K, V>::const_iterator::operator++(int) -> const_iterator
{
    auto old = *this;
    ++*this;
    return old;
}

template<typename K, typename V>
bool Dict<K, V>::const_iterator::operator==(const const_iterator& other) const
{
    return stack_ == other.stack_;
}

template<typename K, typename V>
bool Dict<K, V>::const_iterator::operator!=(const const_iterator& other) const
{
    return stack_ != other.stack_;
}
template<typename K, typename V>
Dict<K, V>::const_iterator::const_iterator(const Node* start)
        : stack_{start}
{ }

} // end namespace dictionary
